import os

rgbn = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn.tif')
rgbn_suba = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_suba.tif')
rgbn_subb = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_subb.tif')

rgbn_20160101 = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_20160101.tif')
rgbn_20160401 = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_20160401.tif')
rgbn_20160517 = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_20160517.tif')
rgbn_20170203 = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_20170203.tif')

rgbn_time_list = [rgbn_20160101, rgbn_20160401, rgbn_20160517, rgbn_20170203]
